/*
 * Copyright (c) MCST, 2021
 *
 * Test Name: ioctl_fatfs
 *
 * Test is intended to verify FATFS-specific cmds of ioctl():
 * FAT_IOCTL_GET_ATTRIBUTES
 * FAT_IOCTL_SET_ATTRIBUTES
 * FAT_IOCTL_GET_VOLUME_ID
 * VFAT_IOCTL_READDIR_BOTH
 */

#define _GNU_SOURCE

#include <linux/msdos_fs.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/mount.h>

#include "test.h"
#include "tst_device.h"
#include "safe_macros.h"
#include "lapi/fcntl.h"
#include "lapi/loop.h"

#define MNTPOINT "mntpoint"
#define TESTDIR "mntpoint/testdir"
#define TESTFILE "mntpoint/testfile"

#define DIRMODE (S_IRWXU | S_IRWXG | S_IRWXO)
#define FILEMODE (S_IRWXU | S_IRWXG | S_IRWXO)

char *TCID = "ioctl_fatfs";

static int testno;
static char dev_path[1024];
static int dir_fd, file_fd, dev_num, attach_flag, mount_flag;

static void setup(void);
static void cleanup(void);
static void verify_fat_ioctl(void);

struct test_case_t {
	void (*setup) (void);
	void (*test_all) (void);
	void (*cleanup) (void);
	char *desc;
} test_data[] = {
	{ NULL, verify_fat_ioctl, NULL, 
		    "specific FATFS ioctl()"}
,};

int TEST_TOTAL = sizeof(test_data) / sizeof(test_data[0]);

int main(int argc, char *argv[])
{
	int lc;

	tst_parse_opts(argc, argv, NULL, NULL);
	setup();

	for (lc = 0; TEST_LOOPING(lc); ++lc) {
		tst_count = 0;
		for (testno = 0; testno < TEST_TOTAL; ++testno) {
			if (test_data[testno].setup != NULL)
				test_data[testno].setup();
			if (test_data[testno].test_all != NULL)
				test_data[testno].test_all();
			if (test_data[testno].cleanup != NULL)
				test_data[testno].cleanup();
		}
	}

	cleanup();
	tst_exit();
}

static void verify_fat_ioctl(void)
{
	uint32_t id = 0, id2 = 0;
	struct __fat_dirent entry[2];

	TEST(ioctl(file_fd, FAT_IOCTL_GET_ATTRIBUTES, &id));
	if (TEST_RETURN < 0) {
		tst_brkm(TBROK, NULL, "ioctl(FAT_IOCTL_GET_ATTRIBUTES): errno=%d",
			TEST_ERRNO);
	}

	/* Invert archive attribute. */
	id ^= ATTR_ARCH;

	TEST(ioctl(file_fd, FAT_IOCTL_SET_ATTRIBUTES, &id));
	if (TEST_RETURN < 0) {
		tst_brkm(TBROK, NULL, "ioctl(FAT_IOCTL_SET_ATTRIBUTES): errno=%d",
			TEST_ERRNO);
	}
	TEST(ioctl(file_fd, FAT_IOCTL_GET_ATTRIBUTES, &id2));
	if (TEST_RETURN < 0) {
		tst_brkm(TBROK, NULL, "ioctl(FAT_IOCTL_GET_ATTRIBUTES): errno=%d",
			TEST_ERRNO);
	}

	if (id != id2)
		tst_resm(TFAIL, "FAT_IOCTL_GET_ATTRIBUTES: expect %d but got %d",
			       	id, id2);
	else
		tst_resm(TPASS, "FAT_IOCTL_GET_ATTRIBUTES: get expected attribite: %d",
			       	id2);

	id = 0;
	TEST(ioctl(file_fd, FAT_IOCTL_GET_VOLUME_ID, &id));
	if (TEST_RETURN < 0) {
		tst_brkm(TBROK, NULL, "ioctl(FAT_IOCTL_GET_VOLUME_ID): errno=%d",
			TEST_ERRNO);
	}
	if (id != 0)
		tst_resm(TPASS, "FAT_IOCTL_GET_VOLUME_ID: get volume ID: %04x-%04x",
			       	id >> 16, id & 0xFFFF);
	else
		tst_resm(TFAIL, "FAT_IOCTL_GET_VOLUME_ID: get %d!", id);

	for (;;) {
		/* Read next directory entry. */
		TEST(ioctl(dir_fd, VFAT_IOCTL_READDIR_BOTH, entry));
		if (TEST_RETURN < 0) {
			tst_brkm(TBROK, NULL, "ioctl(VFAT_IOCTL_READDIR_BOTH): errno=%d",
				TEST_ERRNO);
			break;
		}
		/*
		* If an error occurs, the return value is -1.
		* If the end of the directory list has been reached,
		* the return value is 0.
		* For backward compatibility the end of the directory
		* list is also signaled by d_reclen == 0.
		*/
		if (TEST_RETURN == 0) {
			tst_resm(TPASS, "VFAT_IOCTL_READDIR_BOTH:"
					" read the end of directory list");
			break;
		}
	}
}

static void setup(void)
{
	char *tmpdir;
	tst_tmpdir();
	tmpdir = tst_get_tmpdir();
	tst_resm(TINFO, "tmp dir: %s - OK", tmpdir);
	free(tmpdir);
	dev_num = tst_find_free_loopdev(dev_path, sizeof(dev_path));
	if (dev_num < 0)
		tst_brkm(TBROK, cleanup, "No free loop device");
	tst_fill_file("test.img", 0, 1024 * 1024, 10);
	tst_attach_device(dev_path, "test.img");
	attach_flag = 1;
	tst_mkfs(cleanup, dev_path, "vfat", NULL, NULL);
	tst_resm(TINFO, "mkfs.fat %s - OK", dev_path);
	SAFE_MKDIR(cleanup, MNTPOINT, DIRMODE);
	SAFE_MOUNT(cleanup, dev_path, MNTPOINT, "vfat", 0, NULL);
	mount_flag = 1;
	tst_resm(TINFO, "mount %s %s - OK", dev_path, MNTPOINT);
	SAFE_MKDIR(cleanup, TESTDIR, DIRMODE);
	dir_fd = SAFE_OPEN(cleanup, TESTDIR, O_DIRECTORY);
	file_fd = SAFE_OPEN(cleanup, TESTFILE,
				O_RDWR | O_CREAT, FILEMODE);

	TEST_PAUSE;
}

static void cleanup(void)
{
	if (dir_fd > 0 && close(dir_fd) < 0)
		tst_resm(TWARN | TERRNO, "close newdirfd failed");

	if (file_fd > 0 && close(file_fd) < 0)
		tst_resm(TWARN | TERRNO, "close filefd failed");

	if (mount_flag && tst_umount(MNTPOINT) < 0)
		tst_resm(TWARN | TERRNO, "umount %s failed", MNTPOINT);

	if (attach_flag)
		tst_detach_device(dev_path);

	tst_rmdir();
}

