/*
 * Copyright (c) MCST, 2021
 *
 * Test Name: sockioctl02
 *
 * Test is intended to verify some cmd of ethtool ioctl().
 * SYNOPSYS:
 *     ./sockioctl <optional device list>
 * For example:
 *     ./sockioctl eth0 eth1 eth2
 * Default device list includes single item 'eth0'.
 *
 * NB> Shell wrapper script sockioctl.sh invokes this test in LTP.
 */

#include <stdio.h>
#include <unistd.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>

#include <linux/sockios.h>
#include <linux/ethtool.h>
#include <netinet/in.h>
#include <net/if.h>

#include "test.h"
#include "safe_macros.h"

char *TCID = "sockioctl02";

static int testno;
static int fd;
static struct ifreq ifr;
struct ethtool_pauseparam epause;
struct ethtool_ringparam ering;
struct ethtool_coalesce ecoal;

static void setup(void);
static void setup0(void);
static void setup1(void);
static void setup2(void);

static int check_ring(void);

static void cleanup(void);
static void cleanup0(void);

struct test_case_t {
	int domain;		/* AF_INET, AF_NETLINK */
	int type;		/* SOCK_DGRAM, SOCK_RAW */
	int proto;		/* protocol number */
	int cmd;		/* SIOC* */
	void *arg;
	char *devname; 		/* lo, eth0 */
	int retval;		/* return value */
	int experrno;		/* expected errno */
	void (*setup) (void);
	int  (*test_all) (void);
	void (*cleanup) (void);
	char *desc;
} test_data[] = {
	{
	AF_INET, SOCK_DGRAM, 0, SIOCETHTOOL, &epause, "lo", -1,
		    EOPNOTSUPP, setup0, NULL, cleanup0, 
		    "ethtool -a lo: operation not supported."}
	, {
	AF_INET, SOCK_DGRAM, 0, SIOCETHTOOL, &ering, "eth0", 0,
		    0, setup1, check_ring, cleanup0, 
		    "ethtool -g " /*"eth0"*/}
	, {
	AF_INET, SOCK_DGRAM, 0, SIOCETHTOOL, &ecoal, "eth0", 0,
		    0, setup2, NULL, cleanup0, 
		    "ethtool -c " /*"eth0"*/}
,};

int TEST_TOTAL = sizeof(test_data) / sizeof(test_data[0]);

int main(int argc, char *argv[])
{
	int lc, arg_num = 0;

	tst_parse_opts(argc, argv, NULL, NULL);
	setup();

	for (lc = 0; TEST_LOOPING(lc); ++lc) {
		tst_count = 0;
arg_loop:
		arg_num++;
		if (arg_num >= argc)
			break;
		if (argc > 1) {
			test_data[1].devname = argv[arg_num];
			test_data[2].devname = argv[arg_num];
		}
		for (testno = 0; testno < TEST_TOTAL; ++testno) {
			if ((testno == 0) && (arg_num > 1))
				continue; /* "lo" device has already been tested */
			test_data[testno].setup();
			strcpy(ifr.ifr_name, test_data[testno].devname);

			TEST(ioctl(fd, test_data[testno].cmd, (void*)&ifr));
			if (TEST_RETURN != test_data[testno].retval ||
			    (TEST_RETURN < 0 &&
			     TEST_ERRNO != test_data[testno].experrno)) {
				tst_resm(TFAIL, "%s%s ; returned"
					 " %ld (expected %d), errno %d (expected"
					 " %d)", test_data[testno].desc, test_data[testno].devname,
					 TEST_RETURN, test_data[testno].retval,
					 TEST_ERRNO, test_data[testno].experrno);
				test_data[testno].cleanup();
				continue;
			}
			if (test_data[testno].test_all != NULL) {
				TEST(test_data[testno].test_all());
				if (TEST_RETURN < 0) {
					tst_resm(TFAIL,
						 "CHECK: unexpected returned parameters");
					test_data[testno].cleanup();
					continue;
				}
			}
			tst_resm(TPASS, "%s%s successful",
					 test_data[testno].desc, test_data[testno].devname);
			test_data[testno].cleanup();
		}
		goto arg_loop;
	}

	cleanup();
	tst_exit();
}

static void setup(void)
{
	TEST_PAUSE;
	tst_tmpdir();
}

static void cleanup(void)
{
	tst_rmdir();
}

static void cleanup0(void)
{
	(void)close(fd);
	fd = -1;
}

static void setup0(void)
{
	fd = SAFE_SOCKET(cleanup, test_data[testno].domain, test_data[testno].type,
			test_data[testno].proto);
	memset(&ifr, 0, sizeof(ifr));
	memset(&epause, 0, sizeof(epause));
	epause.cmd = ETHTOOL_GPAUSEPARAM;
	ifr.ifr_data = (void *)&epause;
}

static void setup1(void)
{
	fd = SAFE_SOCKET(cleanup, test_data[testno].domain, test_data[testno].type,
			test_data[testno].proto);
	memset(&ifr, 0, sizeof(ifr));
	memset(&ering, 0, sizeof(ering));
	ering.cmd = ETHTOOL_GRINGPARAM;
	ifr.ifr_data = (void *)&ering;
}

static void setup2(void)
{
	fd = SAFE_SOCKET(cleanup, test_data[testno].domain, test_data[testno].type,
			test_data[testno].proto);
	memset(&ifr, 0, sizeof(ifr));
	memset(&ecoal, 0, sizeof(ecoal));
	ecoal.cmd = ETHTOOL_GCOALESCE;
	ifr.ifr_data = (void *)&ecoal;
}

static int check_ring(void)
{
	struct ethtool_ringparam *er = test_data[testno].arg;

	if (er->rx_pending > 4096)
		return -1;
	if (er->rx_mini_pending > 4096)
		return -1;
	if (er->rx_jumbo_pending > 4096)
		return -1;
	if (er->tx_pending > 4096)
		return -1;
	return 0;
}
