#!/bin/bash

###############################################################################
# Variables
###############################################################################
progname=$(basename $0)

CWD=$(dirname $0)
cd $CWD
CWD=$PWD

VERBOSE=0


GEN_ETH=""
IN_ETH=""
OUT_ETH=""
RCV_ETH=""

PSIZE=64
PCOUNT=10000
PDELAY=1000000

GEN_IP="10.10.1.2"
IN_IP="10.10.1.1"
OUT_IP="10.10.2.1"

UDP_DST_MAX=""
PAUSE=""

###############################################################################
# Print message
###############################################################################
doEcho()
{
	echo "=== ${progname} === $*"
}

verboseEcho()
{
	[ $VERBOSE == 1 ] && doEcho $*
}


###############################################################################
# Show usage and exit
###############################################################################
usage()
{
	doEcho "Usage: $progname -g gen -i in -o out -r recv [-s psize] [-c pcount] [-d pdelay] [-u udp_dst_max] [-p pause] [-hv]"
	exit 1;
}


###############################################################################
# Parsing input parameters and set variables
###############################################################################
while getopts ":g:i:o:r:s:c:d:u:p:hv" optname
do
	case "$optname" in
    "g")
		GEN_ETH=$OPTARG
	;;
	"i")
		IN_ETH=$OPTARG
	;;
    "o")
		OUT_ETH=$OPTARG
	;;
    "r")
		RCV_ETH=$OPTARG
	;;

    "s")
		PSIZE=$OPTARG
	;;
    "c")
		PCOUNT=$OPTARG
	;;
    "d")
		PDELAY=$OPTARG
	;;
    "u")
		UDP_DST_MAX=$OPTARG
	;;
    "p")
		PAUSE=$OPTARG
	;;
	"h")
		usage
	;;
	"v")
		VERBOSE=1
	;;
	"?")
		doEcho "Unknown option $OPTARG"
		usage
	;;
	":")
		doEcho "No argument value for option $OPTARG"
		usage
	;;
	*)
		usage
	;;
	esac
done


###############################################################################
# Run test 
###############################################################################
doEcho "router test case start at $(date)"

verboseEcho "GEN ETH - $GEN_ETH"
verboseEcho "IN ETH - $IN_ETH"
verboseEcho "OUT ETH - $OUT_ETH"
verboseEcho "RCV ETH - $RCV_ETH"
verboseEcho "PSIZE - $PSIZE bytes"
verboseEcho "PCOUNT - $PCOUNT"
verboseEcho "PDELAY - $PDELAY nanoseconds"

[ ! "$GEN_ETH" -o ! "$IN_ETH" -o ! "$OUT_ETH" -o ! "$RCV_ETH" ] && usage

RCV_IP="${OUT_IP%.*}.100"
DST_IP="${OUT_IP%.*}.200"

doEcho "Initializing gen iface $GEN_ETH ..."
ifconfig $GEN_ETH down
ifconfig $GEN_ETH $GEN_IP netmask 255.255.255.0 up

doEcho "Initializing input iface $IN_ETH ..."
ifconfig $IN_ETH down
ifconfig $IN_ETH $IN_IP netmask 255.255.255.0 up

doEcho "Initializing output iface $OUT_ETH ..."
ifconfig $OUT_ETH down
ifconfig $OUT_ETH $OUT_IP netmask 255.255.255.0 up

doEcho "Initializing receive iface $RCV_ETH ..."
ifconfig $RCV_ETH down
ifconfig $RCV_ETH $RCV_IP netmask 255.255.255.0 up

# Pause
[ ! "$PAUSE" == "" ] && sleep $PAUSE

# Wait for ifaces up
while (true);
do
    sleep 1
    ethtool $GEN_ETH | grep "Link detected" | grep -q yes || continue
    ethtool $IN_ETH  | grep "Link detected" | grep -q yes || continue   
    ethtool $OUT_ETH | grep "Link detected" | grep -q yes || continue
    ethtool $RCV_ETH | grep "Link detected" | grep -q yes || continue
    break
done

ip route del ${GEN_IP%.*}.0/24 dev $GEN_ETH 
ip route del ${IN_IP%.*}.0/24 dev $IN_ETH 
ip route del ${OUT_IP%.*}.0/24 dev $OUT_ETH 
ip route del ${RCV_IP%.*}.0/24 dev $RCV_ETH 

ip route add ${OUT_IP%.*}.0/24 dev $OUT_ETH

# Get MAC for IN and RCV ifaces
IN_MAC=$(ip link show $IN_ETH | grep link/ether | awk '{print $2}')
RCV_MAC=$(ip link show $RCV_ETH | grep link/ether | awk '{print $2}')

# Setup forwarding
echo 1 > /proc/sys/net/ipv4/ip_forward

# Setup TTL=1 for OUT iface
iptables -t mangle -A POSTROUTING -o $OUT_ETH -j TTL --ttl-set 1

# Setup ARP record for DST_IP
arp -i $OUT_ETH -s $DST_IP $RCV_MAC

# Accept packets with local source addresses
echo 1 > /proc/sys/net/ipv4/conf/$IN_ETH/accept_local

modprobe pktgen

# Setup generator
cd /proc/net/pktgen
echo "rem_device_all" > ./kpktgend_0
echo "add_device $GEN_ETH" > ./kpktgend_0

echo "count $PCOUNT" > ./$GEN_ETH
echo "delay $PDELAY" > ./$GEN_ETH
echo "pkt_size $PSIZE" > ./$GEN_ETH
echo "dst $DST_IP" > ./$GEN_ETH
echo "dst_mac $IN_MAC" > ./$GEN_ETH
[ ! "$UDP_DST_MAX" == "" ] && echo "udp_dst_max $UDP_DST_MAX" > ./$GEN_ETH

tps_prev=$(cat /sys/class/net/$OUT_ETH/statistics/tx_packets)
echo start > ./pgctrl
tps_curr=$(cat /sys/class/net/$OUT_ETH/statistics/tx_packets)
tps_diff=$(($tps_curr - $tps_prev))
[ $tps_diff -lt $PCOUNT ] && res=FAIL || res=PASS

doEcho "$res : count of routed packets - $tps_diff"

###############################################################################
# Finish test
###############################################################################
doEcho "router test case stop at $(date)"

exit 0 
