#!/bin/bash

###############################################################################
# Variables
###############################################################################
progname=$(basename $0)

CWD=$(dirname $0)
cd $CWD
CWD=$PWD

VERBOSE=0

LTPROOT="/opt/mcst/check/OSLTC/etests/ltp_tc/root"
SCEN="$LTPROOT/runtest/syscalls"
BIN="$LTPROOT/testcases/bin"

TRACE_SYSCALLS=0
TCOUNT=0
FLIST=""
TLIST=""

FAIL=0

###############################################################################
# Print message
###############################################################################
doEcho()
{
	echo "=== ${progname} === $*"
}

verboseEcho()
{
	[ $VERBOSE == 1 ] && doEcho $*
    return 0
}


###############################################################################
# Show usage and exit
###############################################################################
usage()
{
    doEcho "Usage: $progname [-c tcount] [-s] [-hv]"
	exit 1;
}


###############################################################################
# Trace sched_switch
###############################################################################
trace_sched_switch()
{
    doEcho "Trace sched_switch()"

    rm -rf /tmp/lttng_tc-trace-sched_switch

    lttng create lttng_tc-session --output=/tmp/lttng_tc-trace-sched_switch > /dev/null 2>&1
    [ $? != "0" ] && doEcho "FAIL" && return 1

    lttng enable-event --kernel sched_switch
    [ $? != "0" ] && doEcho "FAIL" && return 1

    lttng start > /dev/null 2>&1
    [ $? != "0" ] && doEcho "FAIL" && return 1

    sleep 1
    
    lttng stop > /dev/null 2>&1
    [ $? != "0" ] && doEcho "FAIL" && return 1

    lttng destroy > /dev/null 2>&1
    [ $? != "0" ] && doEcho "FAIL" && return 1

    babeltrace /tmp/lttng_tc-trace-sched_switch 2>&1 | grep -q "sched_switch"
    [ $? != "0" ] && doEcho "FAIL" && return 1

    babeltrace /tmp/lttng_tc-trace-sched_switch 2>&1 | tail -n 5

    doEcho "PASS"
    return 0
}


###############################################################################
# Set trace list
###############################################################################
set_tlist()
{
    FLIST=$(lttng list --kernel --syscall | grep "\[" | grep "\]" | awk '{print $1}')
    TLIST=$FLIST
    [ "$TCOUNT" != 0 ] && TLIST=$(echo $TLIST | tr ' ' '\n' | head -n $TCOUNT)
}


###############################################################################
# Trace system call
###############################################################################
trace_syscall()
{
    sc=$1

    doEcho "Trace system call : $sc"

    cmd=$(grep "^$sc[0-9]" $SCEN | grep -v "#" | head -n 1)
    cmd=${cmd#* }

    [ "_x${cmd}x_" == "_xx_" ] && verboseEcho "Warning : test case not found" && return 0
    verboseEcho "$cmd"

    rm -rf /tmp/lttng_tc-trace
    rm -f /tmp/lttng_tc.strace

    lttng create lttng_tc-session --output=/tmp/lttng_tc-trace > /dev/null 2>&1
    oldsc=$(echo $FLIST | tr ' ' '\n' | grep $sc | grep old)
    tlist="$sc"
    [ "_x${oldsc}x_" != "_xx_" ] && tlist="$tlist,$oldsc"
    lttng enable-event --kernel --syscall $tlist > /dev/null 2>&1
    lttng start > /dev/null 2>&1

    cd $BIN
    strace ./$cmd > /tmp/lttng_tc.strace 2>&1
    res=$?
    cd $CWD
    lttng stop > /dev/null 2>&1
    lttng destroy > /dev/null 2>&1

    [ "$res" != "0" ] && verboseEcho "Warning : test case failed" && return 0


    grep -q "^$sc(" /tmp/lttng_tc.strace 
    [ "$?" != "0" ] && verboseEcho "Warning : system call was not traced by strace" && return 0

    babeltrace /tmp/lttng_tc-trace 2>&1 | grep -q $sc 

    [ $? != "0" ] && doEcho "FAIL" && return 1

    doEcho "PASS"
    return 0
}

###############################################################################
# Parsing input parameters and set variables
###############################################################################
while getopts ":c:shv" optname
do
	case "$optname" in
	"c")
		TCOUNT=$OPTARG
	;;
	"s")
		TRACE_SYSCALLS=1
	;;
	"h")
		usage
	;;
	"v")
		VERBOSE=1
	;;
	"?")
		doEcho "Unknown option $OPTARG"
		usage
	;;
	":")
		doEcho "No argument value for option $OPTARG"
		usage
	;;
	*)
		usage
	;;
	esac
done


###############################################################################
# Run test 
###############################################################################
doEcho "lttng test case start at $(date)"

if [ "$(uname -p)" == "sun4u" ]; 
then
	export LTTNG_DEBUG_NOCLONE=y
	killall -w lttng-sessiond
fi

trace_sched_switch
[ "$?" != "0" ] && FAIL=1


if [ "$TRACE_SYSCALLS" == "1" ];
then
	set_tlist
	verboseEcho "Trace list :"
	verboseEcho "$TLIST"

	for sc in $TLIST;
	do
	    trace_syscall $sc
	
    	[ "$?" != "0" ] && FAIL=1
	done
fi

###############################################################################
# Finish test
###############################################################################
doEcho "lttng test case stop at $(date)"

exit $FAIL
